"""The Method of Fundamental Solutions for the Helmholtz equation

homogeneous Helmholtz Equation

 (\nabla^2 + k^2) u = 0

point source fundamental solution in 3D
\phi_k(x, z) = e^{i |x - z|}/|x - z|

line source fundamental solution (essentially 2D)
\phi_k(x, z) = H_0^{(1)}(k |x - z|)

approximate solution

v(x) = sum_j c_j \phi_k(x, z_j)

collocation at boundary points x_i

\alpha_i v(x_i) + \beta_j \cdot \nabla v(x_i) = \gamma_i

this is a nonlinear least squares problem
linear parameters c_j
nonlinear parameters z_j

input data 
boundary points x_i
Dirichlet coefficients \alpha_i
Neumann coefficients \beta_i
boundary values \gamma_i

this is a separable nonlinear least squares problem
can be solved using varpro (port)
"""

from numpy import array, zeros, ones, arange, identity
from numpy import newaxis, reshape, concatenate
from numpy import hypot, dot, sum, pi, cos, sin, less
from numpy import random, linalg
from scipy import special
from pylab import plot, pcolor, show, clf, figure, axis
from pylab import linspace, meshgrid
from pylab import xlabel, ylabel, title

def unit_circle(m):
    """construct equidistant points on the unit circle

    returns an array with m columns, one point per colum
    """
    t = 2*pi*arange(m)/m
    x = array([cos(t), sin(t)])
    return x

def circle(xc, r, m):
    return xc[:,newaxis] + r * unit_circle(m)    

def fundsol2d(k, x, z):
    d = x - z
    r = hypot(d[0], d[1])
    y = special.h1vp(0, k*r, 0)
    return y

def grad_fundsol2d(k, x, z):
    d = x - z
    r = hypot(d[0], d[1])
    yr = special.h1vp(0, k*r, 1)
    yx = yr[newaxis,...] * k * d / r[newaxis,...]
    return yx

def y_source(k, x, z):
    d = x - z
    r = hypot(d[0], d[1])
    y = special.yv(0, k*r)
    return y

def grad_y_source(k, x, z):
    d = x - z
    r = hypot(d[0], d[1])
    yr = special.yvp(0, k*r, 1)
    yx = yr[newaxis,...] * k * d / r[newaxis,...]
    return yx

class helmholtz:
    def __init__(self, k, x, alpha, beta, gamma, nu=0):
        self.k = k
        self.m = len(x)
        self.x = x
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.nu = nu

    def sol_eval(self, z, c, x):
        x_ = reshape(x, (2,)+x.shape[1:]+(1,)*len(z.shape[1:]))
        z_ = reshape(z, (2,)+(1,)*len(x.shape[1:])+z.shape[1:])
        phi = fundsol2d(self.k, x_, z_)
        return dot(phi, c)

    def residual(self, z, c):
        A = self.matrix(z, self.x)
        resid = dot(A, c) - self.rhs(z)
        return concatenate((resid.real, resid.imag))

    def matrix(self, z, x):
        alpha_ = self.alpha[:,newaxis]
        beta_ = self.beta[:,:,newaxis]
        x_ = x[:,:,newaxis]
        z_ = z[:,newaxis,:]
        phi = fundsol2d(self.k, x_, z_)
        gphi = grad_fundsol2d(self.k, x_, z_)
        A = alpha_ * phi + sum(beta_ * gphi, 0)
        if self.nu != 0:
            A = concatenate((A, self.nu * identity(A.shape[1])), 0)
        return A

    def rhs(self, z):
        if self.nu != 0:
            f = concatenate((self.gamma, zeros(z.shape[1])))
        else:
            f = self.gamma
        return f


def linls(prob, z0):
    A = prob.matrix(z0, prob.x)
    c0, resids, rank, sv = linalg.lstsq(A, prob.rhs(z0))
    kappa = sv.max()/sv.min()
    print "linls", resids, rank, "%.1e %.1e %.1e" % (sv.max(), sv.min(), kappa)
    return c0

class prob_disc_source(helmholtz):
    def __init__(self, k, z_source, m, R, n):
        self.k = k
        self.z_source = z_source
        self.m = m
        self.R = R
        self.n = n
        self.x = unit_circle(m)
        self.z = R * unit_circle(n)

class prob_disc_source_dir(prob_disc_source):
    def __init__(self, k, z_source, m, R, n):
        prob_disc_source.__init__(self, k, z_source, m, R, n)
        alpha = ones(m)
        beta = zeros((2, m))
        gamma = self.sol(self.x)
        helmholtz.__init__(self, k, self.x, alpha, beta, gamma)

    def sol(self, x):
        return -0.25 * y_source(self.k, x, self.z_source[:,newaxis])

class prob_disc_source_neu(prob_disc_source):
    def __init__(self, k, z_source, m, R, n):
        prob_disc_source.__init__(self, k, z_source, m, R, n)
        alpha = zeros(m)
        beta = self.x
        grad_u = -0.25 * grad_y_source(k, self.x, z_source[:,newaxis])
        gamma = sum(grad_u * beta, 0)
        helmholtz.__init__(self, k, self.x, alpha, beta, gamma)

def prob_1(prob, k):
    n = 80
    m = 240
    R = 1.4
    rho = 1.2
    z_source = array([rho, 0])
    return prob(k, z_source, m, R, n)

def prob_disc_source_plot(prob, z, c):
    nc = 200
    R = prob.R
    k = prob.k
    xc = linspace(-R, R, nc)
    yc = linspace(-R, R, nc)
    Xc, Yc = meshgrid(xc, yc)
    Uc = prob.sol_eval(z, c, array([Xc, Yc]))
    Uec = -0.25*y_source(k, prob.z_source[:,newaxis,newaxis], 
                         array([Xc, Yc]))
    print abs(Uc - Uec).max()
    print abs((Uc - Uec)*less(Xc**2+Yc**2, 1)).max()
    vals = [ (1, r"Re $u^{(N)}$", Uc.real, 1.0),
             (2, r"Re $(u^{(N)} - u)$", (Uc - Uec).real, 3e4) ]
    for fig, label, val, scale in vals:
        figure(fig)
        plot_colormap(Xc, Yc, val, -0.2/scale, 0.2/scale)
        plot_circles(prob.x, z)
        rho = hypot(prob.z_source[0], prob.z_source[1])
        title(r"%s for k=%d, R=%g, $\rho=%g$" % (label, k, R, rho))

def plot_colormap(Xc, Yc, val, vmin, vmax):
    clf()
    axis('equal')
    pcolor(Xc, Yc, val, 
           shading="flat", vmin=vmin, vmax=vmax)
    xlabel("x")
    ylabel("y")

def plot_circles(x, z):
    bnd = concatenate((x, x[:,:1]), 1)
    plot(bnd[0], bnd[1])
    plot(z[0], z[1], '+')

if __name__ == "__main__":
    k = 8
    prob = prob_1(prob_disc_source_dir, k)
    z = prob.z
    c = linls(prob, z)
    prob_disc_source_plot(prob, z, c)
    show()
