from numpy import array, zeros, ones, arange
from numpy import newaxis, reshape, concatenate
from numpy import hypot, dot, sum, pi, cos, sin, tanh, arctanh, angle
from pylab import plot, pcolor, show, clf, figure, axis
from pylab import linspace, meshgrid
from pylab import xlabel, ylabel, title

from helmholtz_mfs import helmholtz, circle, linls, plot_colormap
from helmholtz_mfs_nonlin import helmholtz_nonlin, nonlinls, seplinls
from helmholtz_mfs_conv import norm2

class prob_discs(helmholtz_nonlin):
    def __init__(self, k, xcs, rs, bcs, ms, Rs, ns):
        self.xcs = xcs
        self.rs = rs
        self.bcs = bcs
        self.ms = ms
        self.Rs = Rs
        self.ns = ns
        xbs = [circle(array(xc), r, m) 
               for (xc, r, m) in zip(xcs, rs, ms)]
        x = concatenate(xbs, -1)
        alpha = concatenate([ ones(m)*bc 
                              for (m, bc) in zip(ms, bcs)], -1)
        #beta = concatenate([ -xb*(1-bc) for (xb, bc) in zip(xbs, bcs) ], -1)
        beta = concatenate([ -xb*(0 if bc else 1) 
                              for (xb, bc) in zip(xbs, bcs) ], -1)
        gamma = concatenate([ ones(m)*bc for bc in bcs ], -1)
        helmholtz_nonlin.__init__(self, k, x, alpha, beta, gamma)

        zs = [ circle(array(xc), R, n) 
               for (xc, R, n) in zip(self.xcs, Rs, ns) ]
        self.z = concatenate(zs, -1)

    def nonlin_pack(self, z, c):
        q = z.copy()
        low = 0
        for i in range(len(self.ns)):
            high = low + self.ns[i]
            ran = slice(low, high)
            xc = self.xcs[i]
            dx = z[0,ran]-xc[0]
            dy = z[1,ran]-xc[1]
            q[0,ran] = arctanh(2.0*hypot(dx, dy)/self.rs[i]-1.0)
            q[1,ran] = angle(dx+1j*dy)
            low = high
        p = helmholtz_nonlin.nonlin_pack(self, q, c)
        return p

    def nonlin_unpack(self, p):
        q, c = helmholtz_nonlin.nonlin_unpack(self, p)
        z = q.copy()
        low = 0
        for i in range(len(self.ns)):
            high = low + self.ns[i]
            ran = slice(low, high)
            xc = self.xcs[i]
            r = (self.rs[i] * (1.0+tanh(q[0,ran]))/2.0)
            t = q[1,ran]
            z[0,ran], z[1,ran] = xc[0] + r * cos(t), xc[1] + r * sin(t)
            low = high
        return z, c

def prob_discs1(k):
    xcs = [[-0.75, 0.75], [0.5, 0.0]]
    rs = [0.1, 0.3]
    bcs = [1, 0]
    m = 160
    ms = [2*m, 2*m]
    Rs = array(rs) * 0.9
    ns = [m, m]
    return prob_discs(k, xcs, rs, bcs, ms, Rs, ns)

def prob_discs2(k):
    xcs = [[-0.75, 0.75], [0.5, 0.0], [0.0, -0.7]]
    rs = [0.1, 0.3, 0.3]
    bcs = [1, 0, 0]
    ms = [50, 50, 50]
    Rs = array(rs) * 0.9
    ns = [20, 20, 20]
    return prob_discs(k, xcs, rs, bcs, ms, Rs, ns)

def prob_discs_plot(prob, z, c):
    nc = 400
    a = 2
    k = prob.k
    xc = linspace(-a, a, nc)
    yc = linspace(-a, a, nc)
    Xc, Yc = meshgrid(xc, yc)
    Uc = prob.sol_eval(z, c, array([Xc, Yc]))
    label = r"Re $u^{(N)}$"
    figure(1)
    plot_colormap(Xc, Yc, Uc.real, -0.2, 0.2)
    plot(prob.x[0], prob.x[1], '.')
    plot(z[0], z[1], '+')
    title(r"%s for k=%d" % (label, k))


if __name__ == "__main__":
    k = 8
    prob = prob_discs1(k)
    #prob.nu = 1e0
    z0 = prob.z.copy()
    c0 = linls(prob, z0)
    print "c0", abs(c0).max(), norm2(c0)
    solver = ["lin", "nonlin", "seplin"][0]
    if solver == "nonlin":
        z, c = nonlinls(prob, z0, c0)
        c1 = linls(prob, z)
    elif solver == "seplin":
        z, c = seplinls(prob, z0, c0)
    else:
        z, c = z0, c0
    print "c", abs(c).max(), norm2(c)
    prob_discs_plot(prob, z, c)
    show()
